"""Prometheus metrics for the tokenizer service."""
# TODO: disable use in development environments to avoid log spam.

from prometheus_client import Counter, Histogram

FUNCTION_LATENCY = Histogram(
    "tokenizer_function_latency_seconds",
    "Function latency (seconds)",
    buckets=[
        0.001,
        0.003,
        0.005,
        0.01,
        0.025,
        0.05,
        0.075,
        0.1,
        0.25,
        0.5,
        0.75,
        1.0,
        2.5,
        5.0,
        7.5,
        10.0,
        15.0,
        20.0,
        25.0,
        30.0,
        float("inf"),
    ],
    labelnames=["function"],
)

HERMES_FAILURE_COUNT = Counter(
    "tokenizer_hermes_failure_count",
    "Number of failures when calling Hermes",
    labelnames=["reason"],
)

HERMES_TEMPLATE_NAME = Counter(
    "tokenizer_hermes_template_name",
    "Count the number of times a specific template name is used.",
    labelnames=["template_name"],
)

TOKEN_METRICS = Histogram(
    "tokenizer_token_metrics",
    "Metrics reflecting how tokens are handled in tokenizer.",
    buckets=[
        1000,
        1500,
        2000,
        2500,
        3000,
        3500,
        4000,
        5000,
        6000,
        7000,
        8000,
        9000,
        10000,
        15000,
        20000,
        25000,
        30000,
        35000,
        40000,
        45000,
        50000,
        60000,
        70000,
        80000,
        90000,
        100000,
        250000,
        500000,
        750000,
        1000000,
        float("inf"),
    ],
    labelnames=["metric_type"],
)

MESSAGE_METRICS = Histogram(
    "tokenizer_message_metrics",
    "Metrics reflecting how messages are handled in tokenizer.",
    buckets=[
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        15,
        20,
        25,
        30,
        35,
        40,
        45,
        50,
        60,
        70,
        80,
        90,
        100,
        150,
        200,
        250,
        300,
        350,
        400,
        450,
        500,
        600,
        700,
        800,
        900,
        1000,
        1250,
        1500,
        1750,
        2000,
        2500,
        3000,
        3500,
        4000,
        4500,
        5000,
        float("inf"),
    ],
    labelnames=["metric_type"],
)



"""Templating engine for rendering Jinja templates."""

import logging

from decorators import monitor
from exceptions import TemplateNotFoundError
from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound
from metrics import HERMES_TEMPLATE_NAME


class TemplatingEngine:
    """Template engine."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Singleton pattern that allows arbitrary arguments."""
        if cls._instance is None:
            cls._instance = super(TemplatingEngine, cls).__new__(cls)
            # Initialize _instance's attributes
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, logger: logging.LoggerAdapter = None):
        """Initialize template engine."""
        if not self._initialized:
            self._default_template = None
            self._logger = logger if logger else logging.getLogger(__name__)
            self._initialized = True

    @monitor
    def render_template(
        self, context: dict = None, raw_template: str = None, template_name: str = None
    ):
        """Render the jinja template with the provided context data."""
        if context is None:
            context = {}

        if raw_template and template_name:
            raise ValueError("Cannot provide both raw_template and template_name.")

        # Template was provided at runtime.
        if raw_template:
            self._logger.info(
                f"### Hermes is using a runtime template: \n\n {raw_template=}"
            )

            tmp_template = self._load_template_from_string(raw_template)
            return tmp_template.render(context)

        # Load template from disk every time.
        if template_name:
            try:
                tmp_template = self._load_template(template_name=template_name)
                return tmp_template.render(context)
            except TemplateNotFound as ex:
                raise TemplateNotFoundError(
                    f"Template not found: {template_name}. {ex=}"
                )

        # If no template is provided or template could not be rendered, fallback to the default.
        if self._default_template is None:
            self._default_template = self._load_template()

        return self._default_template.render(context)

    def _load_template(self, template_name: str = "production.yml.j2") -> Template:
        """Load template from disk."""
        loader = FileSystemLoader(searchpath="./templates")
        env = Environment(loader=loader)
        # TODO: support dynamic routing to different templates.
        template = env.get_template(template_name)
        HERMES_TEMPLATE_NAME.labels(template_name=template_name).inc()
        return template

    def _load_template_from_string(self, raw_template: str):
        """Load template from raw string."""
        return Template(raw_template)
